import {
  parse
} from "@babel/parser";
import traverse from "@babel/traverse";
import generate from "@babel/generator";
import * as t from "@babel/types";
import UglifyJS from "uglify-js";


export default class Deobfuscator {
  constructor(srcCode) {
    this._code = srcCode;
    this._ast = parse(this._code, { sourceType: 'script' });
    this._strProps = '';
    this._srcArray = [];
    this._shuffleIndex = new Number();
    this._shuffleArray = d => {
      let srcArray = this._srcArray;
      d = function (e) {
          for (; --e; srcArray.push(srcArray.shift()))
          ;
        },
        d(++this._shuffleIndex)
    }
  }


  _setShuffleIndex() {
    traverse.default(this._ast, {
      CallExpression: path => {
        if (path.node.arguments[0] && path.node.arguments[0].name === 'b' && (path.node.arguments[1] && path.node.arguments[1].type === 'NumericLiteral')) {
          this._shuffleIndex = parseInt(path.node.arguments[1].value);
        }
      }
    })
  }

  _setSrcArray() {

    traverse.default(this._ast, {
      MemberExpression: path => {
        if (path.node.object.type === 'StringLiteral' && path.node.object.value.includes('{')) {
          this._strProps = path.node.object.value;
        }
      }
    })
    
    this._srcArray = this._strProps.split('{');
    this._setShuffleIndex();
    //console.log(this._shuffleIndex);
    this._shuffleArray();

    //console.log(this._srcArray);
  }

  _replaceExprToStrings() {
    traverse.default(this._ast, {
      CallExpression: path => {
        if ((path.node.callee.name === 'c') && (path.node.arguments.length === 1) && path.node.arguments[0].value.startsWith('0x')) {
          path.replaceWith(t.stringLiteral(this._srcArray[parseInt(path.node.arguments[0].value)]))
        }
      }
    })
  }

  _minifyCode() {
    this._code = UglifyJS.minify(this._code, {
      compress: {
        toplevel: true,
        sequences: true,
        unsafe: true,
        pure_getters: true
      },
      output: {
        beautify: true
      }
    }).code;
  }

  _replaceSequencesToStatements() {
    this._minifyCode();
    this._ast = parse(this._code);
    traverse.default(this._ast, {
      SequenceExpression: path => {
        if (path.parent && !t.isExpressionStatement(path.parent))
          return;
        path.parentPath.replaceWithMultiple(path.node.expressions.map(el => t.expressionStatement(el)))
      }
    })
  }

  deobfuscateCode() {
    this._setSrcArray();
    this._replaceExprToStrings();
    
    this._code = generate.default(this._ast).code;
    this._replaceSequencesToStatements();
    this._code = generate.default(this._ast).code;

    return { 
      code: this._code,
      ast: this._ast
    }
  }
}