import {
  parse
} from "@babel/parser";
import traverse from "@babel/traverse";
import generate from "@babel/generator";
import * as t from "@babel/types";
import UglifyJS from "uglify-js";


export default class RequestsGenerator {
  constructor({
    code,
    ast,
    _cf_chl_opt
  }) {
    this._code = code;
    this._ast = ast;
    this._cf_chl_opt = _cf_chl_opt;

    this._cf_ctx = {

    }
  }

  _findParam(paramName) {
    let result;
    try {
      traverse.default(this._ast, {
        AssignmentExpression: path => {
          if (path.node.left.property && path.node.left.property.name && path.node.left.property.name === paramName) {
            result = path.node.right.value;
          }
        }
      })
    } catch(_) {}
    return result;
  }

  _set_cf_ctx() {
    this._cf_ctx = {
      chReq: this._cf_chl_opt.cType,
      cNounce: this._cf_chl_opt.cNounce,
      cvId: this._cf_chl_opt.cvId,
      chC: this._findParam('chC'),
      chCAS: this._findParam('chCAS'),
      oV: this._findParam('oV'),
      cRq: this._cf_chl_opt.cRq
    }

    //console.log(this._cf_ctx)
  }

  _getRequiredVarsFromScope(fCode, fAst, fPath) {
    let requiredVars = [];
    fAst = parse(fCode);
    try{
      traverse.default(fAst, {
        ExpressionStatement: path => {
          requiredVars.push(...Object.keys(path.scope.globals));
          throw Error;
        }
      })
    } catch(_){}
    
    requiredVars = requiredVars.filter(item => !(item in global));

    const varsDeclarationCode = `var ${requiredVars.join(', ')}\n;`; //[a, b, c] ===> var a, b, c;
    requiredVars.splice(0, 1);

    
    const currentScope = fPath.scope;
    let varsCode = '';
    requiredVars.forEach(el => {
      if (!currentScope.bindings[el]) return;
      const scopeEl = currentScope.bindings[el].scope;
      currentScope.traverse(scopeEl.path.node, {
        AssignmentExpression: path => {
          if (path.node.left.name === el && path.scope === scopeEl) {
            varsCode += generate.default(path.node).code + '\n;'
          }
        }
      })
    })

    varsCode = varsDeclarationCode + varsCode;

    return varsCode;
  }

  _getFunctionEncryption() {
    const opsEncryption = ['<<', '>>', '<<=', '>>=', '&', '|'];
    let opsEncryptionCx = 0;
    let fPath;
    let fAst;
    let vName;
    try {
      traverse.default(this._ast, {
        AssignmentExpression: path => {

          path.scope.traverse(path.node, {
            AssignmentExpression: path => {
              if (opsEncryption.includes(path.node.operator)) ++opsEncryptionCx;
            },
            BinaryExpression: path => {
              if (opsEncryption.includes(path.node.operator)) ++opsEncryptionCx;
            }
          });

          if (opsEncryptionCx > 40) {
            fPath = path;
            fAst = path.node;
            vName = path.node.left.name;
            throw Error;
          }

          opsEncryptionCx = 0;
        }
      })
    } catch(_) {}
    let fCode = generate.default(fAst).code;
    fCode = this._getRequiredVarsFromScope(fCode, fAst, fPath) + `
    ${fCode}
    const fEcnryption = ${vName}.h;
    fEcnryption(JSON.stringify(${JSON.stringify(this._cf_ctx)}));
    `
    return fCode;



  }

  _getFirstPayload() {
    this._set_cf_ctx();
    const payload = `v_${this._cf_chl_opt.cRay}=${eval(this._getFunctionEncryption()).replace("+", "%2b")}`;
    //console.log(payload)
    return payload;
  }

  _getFirstLink() {
    let linkPath = {};
    let linkCode = '';
    traverse.default(this._ast, {
      ExpressionStatement: path => {
        if (path.node.expression.type === 'CallExpression' &&
            path.node.expression.callee.type === 'MemberExpression' &&
            path.node.expression.callee.property.name === 'sendRequest') {
              linkPath = path.getPrevSibling();
              linkCode = generate.default(linkPath.node).code;
            }
      }
    })
    //console.log(this._getRequiredVarsFromScope(linkCode, {}, linkPath.scope.parent.parent.path))
    
    const evalCode = `
      global._cf_chl_opt = JSON.parse(\`${JSON.stringify(this._cf_chl_opt)}\`);
      var d = global;
      ${this._getRequiredVarsFromScope(linkCode, {}, linkPath.scope.parent.parent.path)};
      ${this._getRequiredVarsFromScope(linkCode, {}, linkPath)};
      let linkCode = ${linkCode};
      (function(){return linkCode})()
    `;
    return `https://www.zoominfo.com${eval(evalCode)}`
  }

  generateFirstRequest() {
    return {
      link: this._getFirstLink(),
      payload: this._getFirstPayload()
    }
  }





}